;; Contactless Thermometer
;;
;; Copyright 2020 - By Michael Kohn
;; http://www.mikekohn.net/
;; mike@mikekohn.net
;;
;; CPU: MSP430G2231
;; Melexis MLX90614 Infra Red Thermometer (contactless).
;; SparkFun Serial 7-Segment display.

.include "msp430x2xx.inc"

RAM equ 0x0200
TEMP equ RAM
I2C_ACK equ RAM+2
REGISTER equ RAM+4
I2C_SLAVE_ADDRESS equ RAM+5
DECIMAL equ RAM+6
FRACTION equ RAM+8
DIGIT_0 equ RAM+12
DIGIT_1 equ RAM+13
DIGIT_2 equ RAM+14
DIGIT_3 equ RAM+15
MLX_TEMPERATURE equ RAM+16
RECEIVE_BUFFER equ RAM+32

CONFIG equ 0x01
UART_TX equ 0x10
BUTTON equ 0x20
I2C_SCL equ 0x40
I2C_SDA equ 0x80

;  r4 = Interrupt counter A
;  r5 = Calibrate: count of 4 cycles every 1 / 16 a second.
;  r6 = Calibrate: CPU cycles for every 1 / 9600 a second.
;  r7 =
;  r8 =
;  r9 =
; r10 =
; r11 =
; r12 =
; r13 = temp in main
; r14 = function param
; r15 = function param and return value

.org 0xf800
start:
  ;; Turn off watchdog
  mov.w #(WDTPW|WDTHOLD), &WDTCTL

  ;; Turn off interrupts
  dint

  ;; Setup stack pointer
  mov.w #0x0280, SP

  ;; Set MCLK to 1.436 MHz with DCO
  mov.b #DCO_3, &DCOCTL
  mov.b #RSEL_8, &BCSCTL1
  mov.b #0, &BCSCTL2

  ;; Set SMCLK to 32.768kHz external crystal
  mov.b #XCAP_3, &BCSCTL3

  ;; Setup output pins
  ;; P1.1 = Motion detector
  ;; P1.2 = config0
  ;; P1.3 = config1
  ;; P1.4 = /RST
  ;; P1.5 = /INT
  ;; P1.6 = i2c SCL
  ;; P1.7 = i2c SDA
  mov.b #I2C_SDA|I2C_SCL|UART_TX, &P1DIR
  mov.b #I2C_SDA|I2C_SCL|UART_TX, &P1OUT
  mov.b #BUTTON|CONFIG, &P1OUT
  mov.b #BUTTON|CONFIG, &P1REN
  ;mov.b #0, &P1SEL
  ;mov.b #0, &P1SEL2

  ;; Setup i2c
  ;mov.b #USISWRST, &USICTL0
  ;bis.b #USIPE7|USIPE6|USIMST, &USICTL0
  ;mov.b #USII2C, &USICTL1
  ;mov.b #USIDIV_3|USISSEL_2|USICKPL, &USICKCTL
  ;bis.b #USISCLREL, &USICNT      ; let the clock float (pulled high by res)
  ;bic.b #USISWRST, &USICTL0      ; clear reset

  ;; Setup Timer A
  mov.w #32768, &TACCR0
  mov.w #TASSEL_1|ID_0|MC_1, &TACTL ; ACLK, DIV1, COUNT
  mov.w #CCIE, &TACCTL0
  mov.w #0, &TACCTL1

  ;; Turn on interrupts
  eint

  ;; To help debugging, clear out some RAM.
  mov.w #0x200, r15
memset:
  mov.w #0, 0(r15)
  add.w #2, r15
  cmp.w #0x230, r15
  jnz memset

  call #delay_1s

  call #calibrate_uart
  call #display_temperature
  call #clear_display

main:
  bit.b #BUTTON, &P1IN
  jnz main
  call #mlx_get_temperature
  call #display_temperature
  call #delay_6s
  call #clear_display
  jmp main

;; mlx_get_id()
mlx_get_id:
  mov.b #0x3c, r15
  mov.b #0x5a, r14
  mov.b #0x01, r13
  call #i2c_read
  ret

;; clear_display()
clear_display:
  mov.b #0x76, r15
  call #send_uart
  ret

;; display_temperature()
display_temperature:
  mov.b #' ', &DIGIT_0
  mov.b #0, &DIGIT_1
  mov.b #0, &DIGIT_2
  mov.b #0, &DIGIT_3
  bit.b #CONFIG, &P1IN
  jnz display_hex
  mov.w &MLX_TEMPERATURE, r15
  call #convert_decimal
  sub.w #273, r14
  mov.w r14, &DECIMAL
  mov.w r15, &FRACTION
  cmp.w #100, r14
  jlo display_temperature_under_100
  mov.b #1, &DIGIT_0
  sub.w #100, r13
display_temperature_under_100:
  mov.w &DECIMAL, r15
  call #divide_10
  mov.b r14, &DIGIT_1
  mov.b r15, &DIGIT_2
  mov.b &FRACTION, r15
  add.w r15, r15
  call #divide_10
  mov.b r14, &DIGIT_3
  ;; Now display.
  mov.b &DIGIT_0, r15
  call #send_uart
  mov.b &DIGIT_1, r15
  call #send_uart
  mov.b &DIGIT_2, r15
  call #send_uart
  mov.b &DIGIT_3, r15
  call #send_uart
  ;; Display decimal point.
  mov.b #0x77, r15
  call #send_uart
  mov.b #0x04, r15
  call #send_uart
  ret

;; convert_decimal(data=r15) = r14=decimal,r15=fraction/50
convert_decimal:
  mov.w #0, r14
convert_decimal_loop:
  cmp.w #50, r15
  jlo convert_decimal_exit
  sub.w #50, r15
  inc.w r14
  jmp convert_decimal_loop
convert_decimal_exit:
  ret

;; divide_10(data=r15) = r14=answer,r15=remainder
divide_10:
  mov.w #0, r14
divide_10_loop:
  cmp.w #10, r15
  jlo divide_10_exit
  sub.w #10, r15
  inc.w r14
  jmp divide_10_loop
divide_10_exit:
  ret

;; display_hex()
display_hex:
  mov.b &MLX_TEMPERATURE+1, r15
  rra.w r15
  rra.w r15
  rra.w r15
  rra.w r15
  mov.b #0x71, r14
  call #i2c_write_byte
  mov.b &MLX_TEMPERATURE+1, r15
  bic.b #0xf0, r15
  mov.b #0x71, r14
  call #i2c_write_byte
  mov.b &MLX_TEMPERATURE, r15
  rra.w r15
  rra.w r15
  rra.w r15
  rra.w r15
  mov.b #0x71, r14
  call #i2c_write_byte
  mov.b &MLX_TEMPERATURE, r15
  bic.b #0xf0, r15
  mov.b #0x71, r14
  call #i2c_write_byte
  ret

;; mlx_get_temperature()
mlx_get_temperature:
  mov.b #0x07, r15
  mov.b #0x5a, r14
  mov.b #0x03, r13
  call #i2c_read
  mov.w &RECEIVE_BUFFER, &MLX_TEMPERATURE
  ret

;; SDA goes low while SCL is still high then SCL goes low.
i2c_start:
  bis.b #I2C_SDA, &P1OUT
  bis.b #I2C_SCL, &P1OUT
  nop
  nop
  nop
  nop
  bic.b #I2C_SDA, &P1OUT
  nop
  nop
  nop
  nop
  bic.b #I2C_SCL, &P1OUT
  ret

;; SCL needs to go high and then SDA goes high.
i2c_stop:
  bis.b #I2C_SCL, &P1OUT
  nop
  nop
  nop
  nop
  bis.b #I2C_SDA, &P1OUT
  ret

;; i2c_write_byte(i2c_address=r14, register=r15)
i2c_write_byte:
  mov.b r15, &REGISTER
  mov.b r14, &I2C_SLAVE_ADDRESS
  rla.w r14
  call #i2c_start
  mov.b r14, r15
  call #i2c_clock_out
  call #i2c_slave_ack
  mov.b &REGISTER, r15
  call #i2c_clock_out
  call #i2c_slave_ack
  call #i2c_stop
  ret

;; i2c_write(i2c_address=r14, register=r15, data=r13)
i2c_write:
  mov.b r15, &REGISTER
  mov.b r14, &I2C_SLAVE_ADDRESS
  rla.w r14
  call #i2c_start
  mov.b r14, r15
  call #i2c_clock_out
  call #i2c_slave_ack
  mov.b &REGISTER, r15
  call #i2c_clock_out
  call #i2c_slave_ack
  mov.b r13, r15
  call #i2c_clock_out
  call #i2c_slave_ack
  call #i2c_stop
  ret

;; i2c_read(i2c_address=r14, register=r15, count=r13)
i2c_read:
  ; First send the a sequence of start, slave address, 0, ack, register, ack.
  mov.b r14, &I2C_SLAVE_ADDRESS
  mov.b r15, &REGISTER
  call #i2c_start
  mov.b &I2C_SLAVE_ADDRESS, r15
  rla.w r15
  call #i2c_clock_out
  call #i2c_slave_ack
  mov.b &REGISTER, r15
  call #i2c_clock_out
  call #i2c_slave_ack
  call #i2c_start
  mov.b &I2C_SLAVE_ADDRESS, r15
  rla.w r15
  bis.b #1, r15
  call #i2c_clock_out
  call #i2c_slave_ack
  ; Next send the a sequence of start, slave address, 1, ack [ read, nack ]
  mov.w #RECEIVE_BUFFER, r12
i2c_read_loop:
  call #i2c_clock_in
  mov.b r15, @r12
  inc.w r12
  call #i2c_master_ack
  dec.b r13
  jnz i2c_read_loop
  call #i2c_stop
  ret

i2c_clock_out:
  mov.b #8, r14
i2c_clock_out_next:
  bic.b #I2C_SDA, &P1OUT
  bit.b #0x80, r15
  jz i2c_clock_out_zero
  bis.b #I2C_SDA, &P1OUT
i2c_clock_out_zero:
  bis.b #I2C_SCL, &P1OUT
  nop
  nop
  nop
  nop
  bic.b #I2C_SCL, &P1OUT
  rla.b r15
  dec.b r14
  jnz i2c_clock_out_next
  bic.b #I2C_SDA, &P1OUT
  ret

i2c_clock_in:
  mov.b #8, r14
  mov.b #0, r15
  bic.b #I2C_SDA, &P1DIR
i2c_clock_in_next:
  rla.w r15
  bis.b #I2C_SCL, &P1OUT
  nop
  nop
  nop
  nop
  bit.b #I2C_SDA, &P1IN
  jz i2c_clock_in_zero
  bis.b #1, r15
i2c_clock_in_zero:
  bic.b #I2C_SCL, &P1OUT
  dec.b r14
  jnz i2c_clock_in_next
  bis.b #I2C_SDA, &P1DIR
  ret

i2c_slave_ack:
  bic.b #I2C_SDA, &P1DIR
  bis.b #I2C_SCL, &P1OUT
  nop
  nop
  nop
  nop
  mov.b #0, r15
  bit.b #I2C_SDA, &P1IN
  jz i2c_slave_ack_off
  mov.b #1, r15
i2c_slave_ack_off:
  bic.b #I2C_SCL, &P1OUT
  bis.b #I2C_SDA, &P1DIR
  ret

i2c_master_ack:
  bic.b #I2C_SDA, &P1OUT
  cmp.b #1, r13
  jnz i2c_master_ack_not_nack
  bis.b #I2C_SDA, &P1OUT
i2c_master_ack_not_nack:
  bis.b #I2C_SCL, &P1OUT
  nop
  nop
  nop
  nop
  bic.b #I2C_SCL, &P1OUT
  ret

send_uart:
  mov.w #8, r14
  dint
  mov.w r6, &TACCR0
  bis.w #TACLR, &TACTL
  mov.w #TASSEL_2|ID_0|MC_1, &TACTL ; SMCLK, DIV1, COUNT
  eint
  ;; Start bit = 0.
  bic.b #UART_TX, &P1OUT
  mov.b #0, r4
send_uart_wait_start_bit:
  cmp.w #1, r4
  jnz send_uart_wait_start_bit
  ;; Send data bits.
send_uart_next_bit:
  rra.w r15
  jnc send_uart_next_0
  bis.b #UART_TX, &P1OUT
  jmp send_uart_bit_set
send_uart_next_0:
  bic.b #UART_TX, &P1OUT
send_uart_bit_set:
  mov.b #0, r4
send_uart_wait_data_bit:
  cmp.w #1, r4
  jnz send_uart_wait_data_bit
  dec.w r14
  jnz send_uart_next_bit
  ;; Stop bit = 1.
  bis.b #UART_TX, &P1OUT
  mov.b #0, r4
send_uart_wait_stop_bit:
  cmp.w #1, r4
  jnz send_uart_wait_stop_bit
  ;; Set Timer A back to 32768 crystal
  mov.w #32768, &TACCR0
  mov.w #TASSEL_1|ID_0|MC_1, &TACTL ; ACLK, DIV1, COUNT
  ret

;; The loop takes 4 cycles.  @1.4MHz that should be 350000 inc a second.
;; 350000 / 16 = 21875.
calibrate_uart:
  mov.w #0, r5
  mov.w #2048, &TACCR0
  bis.w #TACLR, &TACTL
  mov.w #TASSEL_1|ID_0|MC_1, &TACTL ; ACLK, DIV1, COUNT
  mov.w #0, r4
calibrate_uart_loop:
  inc.w r5
  cmp.w #1, r4
  jnz calibrate_uart_loop
  mov.w #32768, &TACCR0
  mov.w #TASSEL_1|ID_0|MC_1, &TACTL ; ACLK, DIV1, COUNT
  mov.w #0, r6
  mov.w r5, r15
calibrate_uart_div_loop:
  cmp.w #150, r15
  jlo calibrate_uart_exit
  sub.w #150, r15
  inc.w r6
  jmp calibrate_uart_div_loop
calibrate_uart_exit:
  dec.w r6
  ret

delay_6s:
  mov.w #0, r4
delay_6s_loop:
  cmp.w #6, r4
  jnz delay_6s_loop
  ret

delay_1s:
  bis.w #TACLR, &TACTL
  mov.w #TASSEL_1|ID_0|MC_1, &TACTL ; ACLK, DIV1, COUNT
  mov.w #0, r4
delay_1s_loop:
  cmp.w #1, r4
  jnz delay_1s_loop
  ret

delay_2s:
  mov.w #0, r4
delay_2s_loop:
  cmp.w #2, r4
  jnz delay_2s_loop
  ret

timer_interrupt_a:
  inc.w r4
  reti

.org 0xfff2
  dw timer_interrupt_a     ; Timer_A2 TACCR0, CCIFG
.org 0xfffe
  dw start                 ; Reset

